/*//////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2023 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
///////////////////////////////////////////////////////////////////////// */
#include <amps_impl.h>
#ifndef _WIN32
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
static pthread_mutex_t _amps_atfork_registry_mutex = PTHREAD_MUTEX_INITIALIZER;

struct _amps_atfork_entry
{
  void*                           user_data;
  _amps_atfork_callback_function  callback;
};


static struct _amps_atfork_entry* _amps_atfork_array;
static size_t _amps_atfork_array_size = 0;
static size_t _amps_atfork_array_capacity = 0;

void __amps_atfork_dispatch(int code_)
{
  for (size_t i = 0; i < _amps_atfork_array_size; ++i)
  {
    _amps_atfork_array[i].callback(_amps_atfork_array[i].user_data, code_);
  }
}

void __amps_prefork(void)
{
  pthread_mutex_lock(&_amps_atfork_registry_mutex);
  __amps_atfork_dispatch(0);
}

void __amps_postfork_parent(void)
{
  __amps_atfork_dispatch(1);
  pthread_mutex_unlock(&_amps_atfork_registry_mutex);
}

void __amps_postfork_child(void)
{
  __amps_atfork_dispatch(2);
  pthread_mutex_unlock(&_amps_atfork_registry_mutex);
}

void amps_atfork_init(void)
{
  pthread_mutex_lock(&_amps_atfork_registry_mutex);
  if (!_amps_atfork_array_capacity)
  {
    _amps_atfork_array = malloc(128 * sizeof(struct _amps_atfork_entry));
    _amps_atfork_array_capacity = 128;
    pthread_atfork(__amps_prefork, __amps_postfork_parent, __amps_postfork_child);
  }
  pthread_mutex_unlock(&_amps_atfork_registry_mutex);
}

void amps_atfork_add(void* user_data_, _amps_atfork_callback_function callback_)
{
  pthread_mutex_lock(&_amps_atfork_registry_mutex);

  if (_amps_atfork_array_size == _amps_atfork_array_capacity)
  {
    struct _amps_atfork_entry* pnew = malloc((128 + _amps_atfork_array_capacity) * sizeof(struct _amps_atfork_entry));
    if (pnew)
    {
      if (_amps_atfork_array)
      {
        memcpy(pnew, _amps_atfork_array, _amps_atfork_array_capacity * sizeof(struct _amps_atfork_entry));
        free(_amps_atfork_array);
      }
      _amps_atfork_array = pnew;
      _amps_atfork_array_capacity += 128;
    }
  }
  if (_amps_atfork_array_size < _amps_atfork_array_capacity)
  {
    _amps_atfork_array[_amps_atfork_array_size++] = (struct _amps_atfork_entry) {user_data_, callback_};
  }

  pthread_mutex_unlock(&_amps_atfork_registry_mutex);
}

void amps_atfork_remove(void* user_data_)
{
  pthread_mutex_lock(&_amps_atfork_registry_mutex);
  for(size_t i = 0; i < _amps_atfork_array_size; ++i)
  {
    if (_amps_atfork_array[i].user_data == user_data_)
    {
      if (i+1 != _amps_atfork_array_size)
      {
        _amps_atfork_array[i] = _amps_atfork_array[_amps_atfork_array_size-1];
      }
      --_amps_atfork_array_size;
      break;
    }
  }
  pthread_mutex_unlock(&_amps_atfork_registry_mutex);
}
#else
AMPSDLL void amps_atfork_init(void)
{
}
AMPSDLL void amps_atfork_add(void* vp_, _amps_atfork_callback_function fn_)
{
}
AMPSDLL void amps_atfork_remove(void* vp_)
{
}
#endif

